(function() {

    ////////////////////////////////////////
    // window size
    ////////////////////////////////////////

    var windowSizeInitialized = false;
    var windowWidth;
    var windowHeight;

    ////////////////////////////////////////
    // current state
    ////////////////////////////////////////

    var currentRenderer = null;
    var currentReaderCallbacks = null;
    var lastBookProviderCallbackId = 0;
    var bookProviderCallbacks = {};

    ////////////////////////////////////////
    // web view initialization
    ////////////////////////////////////////

    window.addEventListener("load", function(event) {
        kindleLiteCallbackListener.clientLoaded();
    });

    ////////////////////////////////////////
    // KindleLiteReaderError
    ////////////////////////////////////////

    function KindleLiteReaderError(errorCode, errorMessage) {
        this.errorCode = errorCode;
        this.errorMessage = errorMessage;
    }

    KindleLiteReaderError.prototype.toString = function() {
        return "Reader error: " + (this.errorMessage ? this.errorMessage + " (" + this.errorCode + ")" : this.errorCode);
    }

    ////////////////////////////////////////
    // book provider
    ////////////////////////////////////////

    window.kindleLiteBookProviderResponse = function(requestId, status, data, timerName) {

        // console.log("karamelBookProviderResponseHandler: " +
        //     "requestId = " + requestId + ", " +
        //     "data = " + JSON.stringify(data));

        if (timerName) {
            console.log("PROF:" + JSON.stringify({ type: "timerEnd", timer: timerName }));
        }

        var thisRequestCallbacks = bookProviderCallbacks[requestId];
        if (!thisRequestCallbacks) {
            return;
        }

        delete bookProviderCallbacks[requestId];

        if (status == "success") {
            thisRequestCallbacks.success(data);
        } else {
            var error = new KindleLiteReaderError(data.errorCode, data.errorMessage);
            thisRequestCallbacks.error(error);
        }

    }

    function registerBookProviderCallbacks(successCallback, errorCallback) {

        lastBookProviderCallbackId++;

        bookProviderCallbacks[lastBookProviderCallbackId] = {
            success: successCallback,
            error: errorCallback
        };

        return lastBookProviderCallbackId;

    }

    function bookProviderGetBookMap(newerThan) {

        var dfd = Q.defer();
        var start = window.performance.now();

        var requestId = registerBookProviderCallbacks(
            function(response) {
                var end = window.performance.now();
                // console.log("Metrics: getBookMap(): " + (end - start) + " ms");
                response.bookmap.version = response.version;
                dfd.resolve(response.bookmap);
            },
            function(error) {
                dfd.reject(error);
            });

        window.kindleLiteBookProvider.getBookMap(requestId, newerThan);

        return dfd.promise;

    }

    function bookProviderGetBookPieces(pieceUrls) {

        return Q.all(pieceUrls.map(function(pieceUrl) {

            var dfd = Q.defer();
            var start = window.performance.now();

            var requestId = registerBookProviderCallbacks(
                function(data) {
                    var end = window.performance.now();
                    // console.log("Metrics: getBookPiece(): " + (end - start) + " ms" + ", pieceUrl = " + pieceUrl);
                    dfd.resolve(data);
                },
                function(error) {
                    dfd.reject(error);
                });

            window.kindleLiteBookProvider.getBookPiece(requestId, pieceUrl);

            return dfd.promise;

        }));

    }

    function bookProviderGetResourceData(resourceUrl, dataType) {

        var dfd = Q.defer();
        var start = window.performance.now();

        var requestId = registerBookProviderCallbacks(
            function(data) {
                var end = window.performance.now();
                // console.log("Metrics: getResourceData(): " + (end - start) + " ms" + ", resourceUrl = " + resourceUrl);
                dfd.resolve(data);
            },
            function(error) {
                dfd.reject(error);
            });

        window.kindleLiteBookProvider.getResourceData(requestId, resourceUrl, dataType);

        return dfd.promise;

    }

    ////////////////////////////////////////
    // reader callbacks
    ////////////////////////////////////////

    function createReaderCallbacks() {

        var terminated = false;

         function dispatchCallback(callbackName, callback) {
              if (!terminated && kindleLiteCallbackListener[callbackName]) {
                 callback();
              }
         }

        return {

            pageStartedRendering: function(pageOrPosition, isPage) {
                dispatchCallback('pageStartedRendering', function () {
                    kindleLiteCallbackListener.pageStartedRendering(pageOrPosition, isPage);
                });
            },

            pageFinishedRendering: function(isImageOnly) {
                dispatchCallback('pageFinishedRendering', function() {
                    var posRange = currentRenderer.getPagePositionRange();
                    var pageNumber = currentRenderer.getPageIndex();
                    kindleLiteCallbackListener.pageFinishedRendering(isImageOnly, posRange.start, posRange.end, pageNumber);
                });
            },

            pageFullyRendered: function() {
                dispatchCallback('pageFullyRendered', function() {
                    kindleLiteCallbackListener.pageFullyRendered();
                });
            },

            pageFailedRendering: function(error) {
                dispatchCallback('readerError', function () {
                    if (error instanceof KindleLiteReaderError) {
                        kindleLiteCallbackListener.readerError(error.errorCode, error.errorMessage);
                    } else {
                        kindleLiteCallbackListener.rendererError(error ? error.toString() : null);
                    }
                });
            },

            internalNavigationCallback: function() {
                dispatchCallback('internalNavigation', function () {
                    kindleLiteCallbackListener.internalNavigation();
                });
            },

            deadZoneClickCallback: function() {
                dispatchCallback('deadZoneClick', function() {
                    kindleLiteCallbackListener.deadZoneClick();
                });
            },

            externalLinkClicked: function(url) {
                dispatchCallback('openExternalLink', function () {
                    kindleLiteCallbackListener.openExternalLink(url);
                });
            },

            internalLinkFollowed: function() {
                dispatchCallback('internalLinkFollowed', function () {
                    kindleLiteCallbackListener.internalLinkFollowed();
                });
            },

            attemptedNavigateOutside: function(direction) {
                dispatchCallback('attemptedNavigateOutside', function () {
                    kindleLiteCallbackListener.attemptedNavigateOutside(direction);
                });
            },

            terminate: function() {
                terminated = true;
            },

            imageViewerVisible: function(visible) {
                dispatchCallback('imageViewerVisible', function () {
                    kindleLiteCallbackListener.imageViewerVisible(visible);
                });
            },

            contentSelected: function(content, language) {
                dispatchCallback('contentSelected', function () {
                    kindleLiteCallbackListener.contentSelected(content, language);
                });
            }

        }

    }

    ////////////////////////////////////////
    // isBookOpen()
    ////////////////////////////////////////

    function isBookOpen() {
        return !!currentRenderer;
    }

    ////////////////////////////////////////
    // openBook()
    ////////////////////////////////////////

    function openBook(asin, startPos, nightModeEnabled, accessibilityEnabled, selectionEnabled, isTextSizePreview, screenWidth, screenHeight) {

        closeBook();

        var body = document.getElementsByTagName("body")[0];

        body.innerHTML = "<div id=\"renderRoot\"></div>";

        body.style.backgroundColor = nightModeEnabled ? "black" : "white";

        var renderRoot = document.getElementById("renderRoot");

        // Determine the window size only once with the first book. We need to do it when the
        // system bars are invisible. So we do it only during the first book open when we can
        // be sure the Android system bars are off. We used to determine the size of the screen
        // on every book open, but this was problematic, because sometimes the book is re-opened
        // when the book UI bars are open, in which case the webview is actually smaller, because
        // we don't use the SYSTEM_UI_FLAG_LAYOUT_HIDE_NAVIGATION flag, because it was problematic
        // in some phones (e.g. Samsung S8). Alternatively, we could calculate the webview size
        // by just getting the webview size in Java and dividing by the pixel density, but this
        // method seems safer.
        if (!windowSizeInitialized) {
            windowWidth = renderRoot.offsetWidth;
            windowHeight = renderRoot.offsetHeight;
            windowSizeInitialized = true;
        }

        var canvasWidth = windowWidth;
        var canvasHeight = windowHeight;

        // This is a quick and dirty workaround for sizing and positioning of the canvas for the
        // text size preview screen. We need a better solution, but for now it's just a horrible
        // hack, because I did not want to touch the book fragment.
        if (isTextSizePreview) {

            canvasWidth = screenWidth;
            canvasHeight = screenHeight;

            var style =
                "width: " + screenWidth + "px; " +
                "height: " + canvasHeight + "px; " +
                "left: " + Math.round((windowWidth - screenWidth) / 2) + "px; " +
                "top: " + 0 + "px;"

            body.innerHTML = "<div id=\"renderRoot\" style=\"" + style + "\"></div>";
            renderRoot = document.getElementById("renderRoot");

        }

        var config = {
            w: canvasWidth,
            h: canvasHeight,
            fontRasterMethod: 0,
            imageLoadMethod: 1,
            navAreaWidth: canvasWidth * 0.25, // TODO: make configurable
            navAreaTopOffset: 50, // TODO: make configurable or determine based on the chrome UI
            navAreaBottomOffset: 0,
            extraClickRadius: 15, // In DP
            transitionAxis: 0, // horizontal
            animatePageTransitions: false,
            doubleTapZoomEnabled: false,
            pageSwipeEnabled: true,
            nightModeEnabled: nightModeEnabled,
            imageViewerEnabled: !accessibilityEnabled, // If accessibility is enabled, disable the Image Viewer
            selectionEnabled: selectionEnabled,
            accessibilityEnabled: accessibilityEnabled
        };

        var bookProvider = {
            getBookMap: bookProviderGetBookMap,
            getBookPieces: bookProviderGetBookPieces,
            getResourceData: bookProviderGetResourceData
        };

        var readerCallbacks = createReaderCallbacks();

        // Old renderer terminated in closeBook();
        currentRenderer = new karamel.KaramelRender(
            bookProvider, renderRoot, config,
            startPos, readerCallbacks);

        currentReaderCallbacks = readerCallbacks;

    }

    ////////////////////////////////////////
    // closeBook()
    ////////////////////////////////////////

    function closeBook() {

        if (!isBookOpen()) {
            return;
        }

        var body = document.getElementsByTagName("body")[0];
        body.innerHTML = "";

        currentReaderCallbacks.terminate();
        bookProviderCallbacks = {}; // Get rid of all callback references to renderer
        currentRenderer.terminate();

        currentReaderCallbacks = null;
        currentRenderer = null;
    }

    ////////////////////////////////////////
    // gotoPageAtPosition()
    ////////////////////////////////////////

    function gotoPageAtPosition(pos) {

        if (!isBookOpen()) {
            return;
        }

        currentRenderer.gotoPageAtPosition(pos);

    }

    function gotoPageAtIndex(page) {

        if (!isBookOpen()) {
            return;
        }

        currentRenderer.gotoPageAtIndex(page);

    }

    ////////////////////////////////////////
    // closeImageViewer()
    ////////////////////////////////////////

    function closeImageViewer() {
        if (!isBookOpen()) {
            return;
        }

        currentRenderer.closeImageViewer();
    }

    ////////////////////////////////////////
    // clearSelection()
    ////////////////////////////////////////

    function clearSelection() {
        if (!isBookOpen()) {
            return;
        }

        currentRenderer.clearSelection();
    }

    ////////////////////////////////////////
    // For touch-to-explore selection
    ////////////////////////////////////////
    function getWordAtPoint(x, y) {
        if (!isBookOpen()) {
            return;
        }

        return currentRenderer.getWordAtPoint(x, y);
    }

    function performSelection(x, y) {
        if (!isBookOpen()) {
            return;
        }

        currentRenderer.handleSelectionAtPoint(x,y);
    }

    ////////////////////////////////////////
    // external API's
    ////////////////////////////////////////

    window.closeBook = closeBook;
    window.openBook = openBook;
    window.gotoPageAtPosition = gotoPageAtPosition;
    window.gotoPageAtIndex = gotoPageAtIndex;
    window.closeImageViewer = closeImageViewer;
    window.clearSelection = clearSelection;
    window.getWordAtPoint = getWordAtPoint;
    window.performSelection = performSelection;

})();